//-----------------------------------------------------------------------------
//
//  Audio.c++
//
//  Audio Input and Output Class Implementation
//
//  Tony Jebara (Copyright (C) 1997)
//  Brian Clarkson
//
//-----------------------------------------------------------------------------

#include "Audio.h++"

// Booleans
#define TRUE 1
#define FALSE 0

AudioIn::AudioIn() {
  // Init the state
  IsPortOpen = FALSE;

  // Get the resource id for the mic
  mic = alGetResourceByName(AL_SYSTEM,"microphone",AL_INPUT_DEVICE_TYPE);
  if(!mic) {
    fprintf(stderr,"Can't find microphone!\n");
    exit(1);
  }

  // Get the resource id for the speaker
  speaker = alGetResourceByName(AL_SYSTEM,"speaker",AL_OUTPUT_DEVICE_TYPE);
  if(!speaker) {
    fprintf(stderr,"Can't find speaker!\n");
    exit(1);
  }

  // Init the audio configuration
  audioConfigIn  = alNewConfig();
  audioConfigOut = alNewConfig();
  
  alSetSampFmt(audioConfigIn, AL_SAMPFMT_FLOAT);
  alSetSampFmt(audioConfigOut, AL_SAMPFMT_FLOAT);
  
  // change the number of channels to mono
  alSetChannels(audioConfigIn, AL_MONO);
  alSetChannels(audioConfigOut, AL_MONO);

  // set the input queue size for the audio port
  alSetQueueSize(audioConfigIn, INPUT_Q_SIZE);
  alSetQueueSize(audioConfigOut, INPUT_Q_SIZE);
  // the q size for the circular buffer is by default the same
  qsize = INPUT_Q_SIZE;

  // Set the max and min
  alSetFloatMax(audioConfigIn, 1.0);
  alSetFloatMax(audioConfigOut, 1.0);

  // set input device
  alSetDevice(audioConfigIn, mic);
  alSetDevice(audioConfigOut, speaker);

  // Some global values
  pvs[0].param = AL_RATE;
  pvs[0].value.ll = alDoubleToFixed(SAMPLE_RATE);

}

AudioIn::~AudioIn()
{
  if(IsPortOpen)
    {
      ALcloseport(inputPort);
      ALcloseport(outputPort);
    }
  alFreeConfig(audioConfigIn);
  alFreeConfig(audioConfigOut);
}

void
AudioIn::SetQSize(int s)
{
  qsize = s;
  Reset(qsize);
}

void
AudioIn::Open()
{
  int err;
  int rv;

  // Make sure the port isn't already open
  if(IsPortOpen)
    return;


  // Set the configuration
  if(alSetParams(mic, pvs, 1) < 0)
    printf("Failed to set audio parameters: %s\n",alGetErrorString(oserror()));
  switch(pvs[0].sizeOut) {
  case AL_INVALID_PARAM:
    fprintf(stderr,"Invalid audio parameter while setting sample rate\n");
    break;
  case AL_INVALID_VALUE:
    fprintf(stderr,"Invalid audio value while setting sample rate\n");
    break;
  case 0:
    fprintf(stderr,"Unknown error while setting sample rate\n");
    break;
  }

  // Set the configuration
  if(alSetParams(speaker, pvs, 1) < 0)
    printf("Failed to set audio parameters: %s\n",alGetErrorString(oserror()));
  switch(pvs[0].sizeOut) {
  case AL_INVALID_PARAM:
    fprintf(stderr,"Invalid audio parameter while setting sample rate\n");
    break;
  case AL_INVALID_VALUE:
    fprintf(stderr,"Invalid audio value while setting sample rate\n");
    break;
  case 0:
    fprintf(stderr,"Unknown error while setting sample rate\n");
    break;
  }

  // set the input and output sample rates
  inputPort  = alOpenPort("Cogno", "r", audioConfigIn);
  outputPort = alOpenPort("Speak", "w", audioConfigOut);

  // Some error checking
  if (inputPort == (ALport)0) {
    err = oserror();
    if (err == AL_BAD_NO_PORTS) {
      fprintf(stderr, " System is out of audio ports\n");
    } else if (err == AL_BAD_DEVICE_ACCESS) {
      fprintf(stderr, " Couldn't access audio device\n");
    } else if (err == AL_BAD_OUT_OF_MEM) {
      fprintf(stderr, " Out of memory: port open failed\n");
    }
    exit(1);
  }

  // Some error checking
  if (outputPort == (ALport)0) {
    err = oserror();
    if (err == AL_BAD_NO_PORTS) {
      fprintf(stderr, " System is out of audio ports\n");
    } else if (err == AL_BAD_DEVICE_ACCESS) {
      fprintf(stderr, " Couldn't access audio device\n");
    } else if (err == AL_BAD_OUT_OF_MEM) {
      fprintf(stderr, " Out of memory: port open failed\n");
    }
    exit(1);
  }

  // Set the state
  IsPortOpen = TRUE;

  // reset the circular buffer
  Reset(qsize);
}

void
AudioIn::Close()
{
  if(IsPortOpen)
    {
      alClosePort(inputPort);
      alClosePort(outputPort);
      IsPortOpen = FALSE;
    }
}

int
AudioIn::Read(void)
{
  int len = 0;;

  if(IsPortOpen) {
    len = alGetFilled(inputPort);
    Read(len);
  }

  return len;
}

int
AudioIn::Write(int len, float *buffer)
{
  return(alWriteFrames(outputPort,buffer,len));
}

int
AudioIn::Write(int len, float *buffer, int startfrom)
{
  int fillable;
  int startagain;

  if ((startfrom>=0) && (startfrom<len))
    {
      fillable = alGetFillable(outputPort);
      if (fillable<(len-startfrom))
	{
	  alWriteFrames(outputPort,&buffer[startfrom],fillable);
	  startagain = startfrom+fillable;
	}
      else
	{
	  alWriteFrames(outputPort,&buffer[startfrom],(len-startfrom));
	  startagain = len;
	}
    }
  else
    {
      if (startfrom<0)    startagain = 0;
      if (startfrom>=len) startagain = len;
    }
  return(startagain);
}

void
AudioIn::Read(int len) 
{
  if(IsPortOpen) {
    int n;
    // Detect wrap around in the circular buffer
    n = size - pos; // space left
    if(n < len) {
      // Read what we can
      //fprintf(stderr,"Reading %d samples (size = %d)\n",n,size);
      if(n + pos > size)
	fprintf(stderr,"Out of bounds in audio buffer\n");
      alReadFrames(inputPort,&(data[pos]),n);
      pos = 0;
      // Then wrap around and read the rest
      Read(len - n);
    } else { 
      // Read all the data
      //fprintf(stderr,"Reading %d samples (size = %d)\n",len,size);
      if(len + pos > size)
	fprintf(stderr,"Out of bounds in audio buffer\n");
      alReadFrames(inputPort,&(data[pos]),len);
      // Increment the circular buffer position
      pos += len;
    }  
  } else {
    fprintf(stderr,"Trying to read when the audio port is closed!\n");
  }
}


void AudioIn::MinMax() {
  int saved_size;
  float sum = 0.0;
  float * samples;
  int i;
  
  // Save the old size and sample 10 secs of audio
  saved_size = size;
  Reset(SAMPLE_RATE*10);

  fprintf(stderr,"\n");
  for(i = 0; i < 10; i++) {
    printf(" Sampling for min and max (t = %d) \r",10 - i);
    Read(SAMPLE_RATE);
  }
  fprintf(stderr,"\n");
  
  // Calc mean, min, and max
  samples = GetLast(SAMPLE_RATE*10);
  max = -1.0;
  min = 1.0;
  for(i = 0; i < SAMPLE_RATE*10; i++) {
    sum += samples[i];
    if(samples[i] > max)
      max = samples[i];
    if(samples[i] < min)
      min = samples[i];
  }
  mean = sum/(SAMPLE_RATE*10);

  fprintf(stderr,"\tmean = %f, range = [%f, %f]\n",mean,min,max);

  Reset(saved_size);
}
